div.sphinxsidebar ul li.toctree-l1 {
    list-style-position: outside;
    margin: 4px 0 0 0;
    padding-left: 1.5em;
    text-indent:-1.5em;
}

div.sphinxsidebar ul li.toctree-l1 > a {
    font-size: 100%;
}

div.sphinxsidebar ul li.toctree-l2 > a {
    font-size: 100%;
}

div.sphinxsidebar p.caption {
    font-size: 110%;
    font-weight: bold;
}